/*
 *	init.c
 *
 *	initialization routines
 *
 *	$Id: init.c,v 1.61 2012/03/05 03:59:59 conrads Exp $
 *
 *****************************************************************************/

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "mkreadmes.h"

char	categories[MAXCATLIST];			/* list of categories found in the
											top-level Makefile
										 */
int		num_categories = 0;				/* number of categories found */

extern	char	ports_dir[PATH_MAX];		/* top-level ports directory */
extern	char	templates_dir[PATH_MAX];	/* directory containing the README.*
												template files
											 */
/* pointers to dynamically allocated buffers for the template files */
extern	char	*template_buf_top;
extern	char	*template_buf_category;
extern	char	*template_buf_port;

/* prototypes for routines in other source files */

extern	int		load_index(void);	/* index.c */

/*****************************************************************************/

static void cleanup(void)
/*
	final cleanup routine, to release all of the memory we allocated

	registered with atexit(3) during initialization, to be called
	automatically when program exits

	returns nothing (doesn't return at all, in fact)
*/
{
	extern char **index_categories;
	extern char  *index_buf;

	if (index_buf)
		free(index_buf);

	if (template_buf_top)
		free(template_buf_top);

	if (template_buf_category)
		free(template_buf_category);

	if (template_buf_port)
		free(template_buf_port);

	if (index_categories)
		free(index_categories);
}

/*****************************************************************************/

char *skip_leading(const char *str, const char *pattern)
/*
	skip over the string 'pattern' at the beginning of 'str'

	returns	a pointer to the location in 'str' immediately following 'pattern',
	or the original	pointer if 'str' does not begin with 'pattern'
 */
{
	int len = strlen(pattern);

	return (strncmp(str, pattern, len) == 0) ?
		(char *) (str + len) : (char *) str;	/* using casts here to quiet a
													compiler warning */
}

/*****************************************************************************/

static int gen_categories_list(void)
/*
 *	generate a list of all ports categories from the top-level Makefile
 *
 *	returns number of categories found, or -1 on error
 */
{
	char	Makefile_path[PATH_MAX];
	char	Makefile_buf[LINE_MAX];
	FILE   *Makefile;

	char   *categories_ptr = categories;	/* pointer into categories array */
	char   *category;		/* pointer to category in current input buffer */

	/* generate the pathname for the top-level Makefile */

	strcpy(Makefile_path, ports_dir);
	strcat(Makefile_path, MAKEFILENAME);

	/* open the file for reading */

	if ((Makefile = fopen(Makefile_path, "r")) == NULL)
	{
		perror(Makefile_path);
		return -1;
	}

	/* scan through all the "SUBDIR += " lines to get all the categories */

	while (fgets(Makefile_buf, LINE_MAX, Makefile) != NULL)
	{
		if ((category = skip_leading(Makefile_buf, "SUBDIR += ")) !=
				Makefile_buf)	/* found a SUBDIR line */
		{
			/* don't want the newline */
			category[strlen(category) - 1] = '\0';

			/* copy category to categories[], bump pointer to next free slot */
			categories_ptr = stpcpy(categories_ptr, category) + 1;
			++num_categories;
		}
	}

	/* add an extra nul character to mark the end of the categories array */
	*categories_ptr = '\0';

	fclose(Makefile);
	return 0;
}

/*****************************************************************************/

static char *load_template(char *ext)
/*
	load the template file with extension 'ext' into a dynamically allocated
	buffer

	returns a pointer to the buffer on success, or NULL on failure
 */
{
	char	template_path[PATH_MAX];

	char   *buffer;		/* pointer to memory obtained from malloc(3) */
	int		fd;			/* file descriptor for open() and close() */
	long	size;		/* file's size, as returned by stat(2) */
	struct	stat sb;	/* will hold the results from stat(2) */

	/* generate the fully qualified pathname for the template file */
	strcpy(template_path, templates_dir);
	strcat(template_path, TEMPLATENAME);	/* template name minus the
												extension
											 */
	/* add the extension to the template pathname */
	strcat(template_path, ext);

	/* get the file's size */
	if (stat(template_path, &sb) == -1)
	{
		perror(template_path);
		return NULL;
	}

	size = sb.st_size;

	/* allocate (size + 1) bytes, so we can mark end of file
		with an extra nul character
	 */
	if ((buffer = malloc(size + 1)) == NULL)
	{
		perror("malloc failure");
		return NULL;
	}

	/* open and read the entire file into memory */

	if ((fd = open(template_path, O_RDONLY|O_DIRECT)) == -1)
	{
		perror(template_path);
		return NULL;
	}

	if (read(fd, buffer, size) != size)
	{
		perror(template_path);
		close(fd);
		return NULL;
	}

	/* add an extra nul to mark end of buffer */

	buffer[size] = '\0';

	close(fd);

	return buffer;
}

/*****************************************************************************/

static int load_templates(void)
/*
	load each of the template files into memory

	returns 0 on success, or -1
*/
{
	char *ptr;

	/* if templates_dir was not already set by the user (using the "-t" command
		line option), set it to ${PREFIX}/TEMPLATESDIR
	 */
	if (!strlen(templates_dir))
	{
		/* honor the PREFIX environment variable */
		strcpy(templates_dir, PREFIX);
		strcat(templates_dir, TEMPLATESDIR);
	}
	else if(templates_dir[strlen(templates_dir) - 1] != '/')
		strcat(templates_dir, "/");

	/* load each of the three template files, README.top, REAME.category and
		README.port
	 */
	if ((template_buf_top = load_template("top")) == NULL)
		return -1;
	if ((template_buf_category = load_template("category")) == NULL)
		return -1;
	if ((template_buf_port = load_template("port")) == NULL)
		return -1;

	return 0;
}

/*****************************************************************************/

int initialize()
/*
	Initialize the program by performing the following series of operations:

	Register a cleanup routine to be called automatically when the program
	exits, to ensure that we free up whatever memory we obtain via malloc()

	Unless it was already set by the "-p" command line option, set the
	ports_dir variable to the top-level ports directory, either from the
	environment variable PORTSDIR (if set), or the #defined value of
	PORTSTOP ("/usr/ports/").

	Generate a list of categories from the top-level Makefile.

	Load the index file	into a dynamically allocated buffer.  The
	load_index_file() routine will also generate a list of pointers to the
	start of each category's section within the index buffer, to optimize
	ports searches later

	Load each of the three template files into their own dynamically allocated
	buffer

	Returns 0 on success, or -1 on error
*/
{
	char *ptr;

	/* register a cleanup routine with atexit(3) to automatically
		free whatever memory we allocated when the program exits
	 */
	if (atexit(cleanup) == -1)
	{
		perror("atexit() register of cleanup routine failed");
		return -1;
	}

	/* if ports_dir was not already set by the user (using the "-p" command line
		option), then set it to the current setting of the PORTSDIR environment
		variable (if defined), or the default PORTSTOP
	 */
	if (!strlen(ports_dir))
	{
		if ((ptr = getenv("PORTSDIR")) != NULL)
		{
			strcpy(ports_dir, ptr);

			/* make sure ports_dir ends with a "/" */

			if (ports_dir[strlen(ports_dir) - 1] != '/')
				strcat(ports_dir, "/");
		}
		else
			strcpy(ports_dir, PORTSTOP);
	}
	else if (ports_dir[strlen(ports_dir) - 1] != '/')
	{
			strcat(ports_dir, "/");
	}

	/* generate a list of categories from the top-level Makefile */
	if (gen_categories_list() != 0)
		return -1;

	/* load the index file into memory */
	if (load_index() == -1)
	{
		fprintf(stderr, "Unable to load the index file!\n");
		return -1;
	}

	/* load all of the template files */
	return load_templates();
}

