/*
 *	Program: mkreadmes
 *
 *	A much faster version of "make_readmes", the Perl-based FreeBSD ports tool
 *	used in FreeBSD's ports make system for (re)building the ports README.html
 *	files
 *
 *	Unlike the Perl version, this program is designed to be used as a
 *	standalone tool, not intended for integration into the FreeBSD ports
 *	collection's system of Makefiles (although it very well could be)
 *
 *	Depends on the following files within the ports tree:
 *
 *	the ports index file (INDEX-${OSRELEASE}, e.g., INDEX-10), used to gather
 *	information about a specific port
 *
 *	the top-level and category-level Makefiles (to obtain information about
 *	the existing categories and ports)
 *
 *	the template files under the Templates directory (README.top,
 *	README.category and README.port), used as a "blueprint" for generating
 *	the README.html files
 *
 *	This file contains only the main() function and a few small auxiliary
 *	functions called directly by main(), plus a few global variables.
 *
 *	The three real "workhorse" routines that drive the program --
 *	make_top_readme(), make_category_readme() and make_port_readme(), as well
 *	as the "wrapper" routine make_readme() which is called from main() --
 *	all are located in the mkreadmes.c file.
 *
 *	$Id: main.c,v 1.38 2012/03/05 03:59:59 conrads Exp $
 *
 *****************************************************************************/

#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "mkreadmes.h"

char	ports_dir[PATH_MAX];		/* top-level ports directory */

char	templates_dir[PATH_MAX];	/* directory containing the README.*
										template files
									 */
char	*category = NULL;			/* pointer to current category */

/* flags set by command line options */
static	int	normal_paths = 0;	/* make relative paths relative to current
									working directory instead of ports_dir
								 */
int		quiet = 0;				/* silence normal output */
int		verbose = 0;			/* increase the amount of output */

/* prototypes for routines in other source files */

extern	char	*skip_leading(const char *str, const char *pattern);/* init.c */
extern	int		initialize(void);									/* init.c */
extern	int		make_readme(const char *path, int make_level);	/* mkreadmes.c */

/******************************************************************************/

static int get_make_level(char *path)
/*
	Distinguish whether the fully qualified 'path' is a port, category or TOP
*/
{
	/* Is it the top ports directory? */

	if (strcmp(path, ports_dir) == 0)
	{
		return TOP;
	}

	/* Not TOP.  OK, then, is it a category? */

	/* remove leading ports_dir/ from path
		(setting 'category' now saves us a bit of work later in
		make_readme_category() and make_readme_port())
	 */

	category = skip_leading(path, ports_dir);

	/* if category still contains a slash, then it's not at category level,
		so try it as a port
	 */
	if (strcspn(category, "/") != strlen(category))
	{
		return PORT;
	}

	return CATEGORY;	/* let make_readme_category() figure out if it's a
							valid category or not
						 */
}

/******************************************************************************/

static char *resolve_path(char *path)
/*
	Resolve the pathname pointed to by 'path' to a fully qualified pathname

	As a convenience, let any relative path be relative to ports_dir,
	unless the user selected the "-n" option, in which case path will be
	relative to the current working directory

	Returns a pointer to the fully qualified pathname returned by realpath(3),
	or NULL on error

	Uses realpath(3) to do the grunt work

	The caller should free() the returned pointer when it is no longer needed
*/
{
	char *real_path;				/* result from realpath(3) */
	char pathbuf[PATH_MAX];			/* temporary buffer */
	char *ptr;

	if (path == NULL)				/* use default path (top-level) */
	{
		strcpy(pathbuf, ports_dir);
	}
	else if (*path == '/')			/* absolute path */
	{
		strcpy(pathbuf, path);
	}
	else if (strcmp(path, ".") == 0)	/* current working directory */
	{
		if (getcwd(pathbuf, PATH_MAX) == NULL)
		{
			perror("getcwd");
			return NULL;
		}
	}
	else if (strcmp(path, "..") == 0)	/* parent directory */
	{
		if (getcwd(pathbuf, PATH_MAX) == NULL)
		{
			perror("getcwd");
			return NULL;
		}

		if ((ptr = dirname(pathbuf)) == NULL)
		{
			perror("dirname");
			return NULL;
		}
		else
		{
			strcpy(pathbuf, ptr);
		}
	}
	else							/* relative path */
	{
		if (!normal_paths)			/* append path to ports_dir */
		{
			strcpy(pathbuf, ports_dir);
			strcat(pathbuf, path);
		}
		else					/* append path to current working directory */
		{
			if (getcwd(pathbuf, PATH_MAX) == NULL)
			{
				perror("getcwd");
				return NULL;
			}
			else
			{
				/* make sure pathbuf ends with a "/" before appending to it */

				if (pathbuf[strlen(pathbuf) - 1] != '/')
					strcat(pathbuf, "/");

				strcat(pathbuf, path);
			}
		}
	}

	/* make sure pathbuf ends with a "/" */

	if (pathbuf[strlen(pathbuf) - 1] != '/')
					strcat(pathbuf, "/");

	/* Now run what we have through realpath to "smooth it out" */

	if ((real_path = realpath(pathbuf, NULL)) == NULL)
	{
		perror(pathbuf);
		return NULL;
	}

	return real_path;
}

/******************************************************************************/

static void usage(void)
{
	char usage_text[] =
"\n"
"mkreadmes -- Make README.html files for the FreeBSD Ports Collection\n"
"\n"
"Usage: mkreadmes [options] [pathname-list]\n"
"\n"
"Options:\n"
"-h             Display this help.\n"
"-n             Use normal handling of relative paths, instead of making them\n"
"               relative to ${PORTSDIR} (\".\" and \"..\" are always treated\n"
"               normally).\n"
"-p pathname    Use pathname as the top level of the ports tree.  Any relative\n"
"               paths in pathname-list will be relative to this, unless the \"-n\"\n"
"               switch is used.  Defaults to the setting of the environment\n"
"               variable \"${PORTSDIR}\", if defined, or \"/usr/ports\" otherwise.\n"
"-q             Quiet mode.  Don't output anything except error messages.\n"
"-t pathname    Use the template files stored under pathname instead of the\n"
/* port Makefile will edit the prefix in the next line */
"               default /usr/local/share/mkreadmes/Templates.\n"
"-V             Display program version.\n"
"-v             Verbose mode.  Show more details about the program's progress.\n"
"\n"
"pathname-list is an optional whitespace-separated list of port and category\n"
"pathnames.  If none is provided, default is to make all of the README.html\n"
"files for the entire ports tree (top-level, categories and ports).\n"
"\n"
"\0";

	char *ptr = usage_text;

	while (*ptr)
	{
		printf("%s", ptr);
		ptr += strlen(ptr) + 1;
	}

	exit (0);
}

/******************************************************************************/

static int validate_path(char *path)
/*
	verify that path exists and is a directory

	returns 0 on success, or -1
 */
{
	struct	stat sb;				/* information returned from stat() */

	/* call stat to get info for path */

	if (stat(path, &sb) == -1)
	{
		perror(path);
		return -1;
	}

	if (!S_ISDIR(sb.st_mode))
	{
		fprintf(stderr, "%s is not a directory\n", path);
		return -1;
	}

	return 0;
}

/******************************************************************************/

int main(int argc, char *argv[])
/*
	program entry point

	process each command line argument in turn

	returns the number of command line arguments which resulted in some
	type of error, or -1 if initialization failed
 */
{
	char *path = NULL;	/* current pathname command line argument */

	char c;				/* character returned from getopt() */

	int make_level;		/* flag to indicate the level within the ports tree
							that we're starting from for a given command line
							argument, will be set to one of TOP, CATEGORY or
							PORT for each pathname we process
						 */

	int	return_code = 0;	/* code returned on exit*/

	extern int optind;

#define suggest_help	fprintf(stderr, "Try \"mkreadmes -h\" for help\n")

	/* process command line arguments and options */

	while ((c = getopt(argc, argv, ":hnp:qt:Vv")) != -1)
	{
		switch (c)
		{
			case 'h':
				usage();	/* usage() exits after displaying help */
			case 'n':
				++normal_paths;
				break;
			case 'p':
				strcpy(ports_dir, optarg);
				if (validate_path(ports_dir) != 0)
					++return_code;
				break;
			case 'q':
				if (verbose)
				{
					fprintf(stderr, "option \"-q\" is incompatible with \"-v\"\n");
					++return_code;
				}
				else
				{
					++quiet;
				}
				break;
			case 't':
				strcpy(templates_dir, optarg);
				if (validate_path(templates_dir) != 0)
					++return_code;
				break;
			case 'V':
				printf("mkreadmes version ");
				printf(VERSION);
				printf("\n");
				exit (0);
			case 'v':
				if (quiet)
				{
					fprintf(stderr, "option \"-v\" is incompatible with \"-q\"\n");
					++return_code;
				}
				else
				{
					fprintf(stderr, "Increasing verbosity level\n");
					++verbose;
				}
				break;
			case ':':
				fprintf(stderr, "Missing pathname argument to option \"-%c\"\n",
						optopt);
				suggest_help;
				++return_code;
				break;
			case '?':
				fprintf(stderr, "Unknown option: -%c\n", optopt);
				suggest_help;
				++return_code;
				break;
			default:
				printf ("?? getopt returned character code 0%d ??\n", c);
		}
	}

	if (return_code)		/* errors in command line processing */
		return return_code;

	/* do initial setup */

	if (initialize() == -1)
	{
		return -1;			/* initialization failed */
	}

	/* if no pathnames given as command line arguments,
		use default dir (top-level)
	 */
	if (optind == argc)
	{
		return make_readme(ports_dir, TOP);
	}

	/* otherwise, process each command line argument in turn */

	while (optind < argc)
	{
		if ((path = resolve_path(argv[optind++])) == NULL)
		{
			++return_code;	/* note the error and move on to next argument */
			continue;
		}

		make_level = get_make_level(path);	/* will ALWAYS be one of TOP,
												CATEGORY or PORT; no need to
												check for error
											 */

		return_code += make_readme(path, make_level);

		/* free the pointer malloc'ed by realpath(3) earlier */
		free(path);
	}

	return return_code;
}

