/*
 *	File: mkreadmes.c
 *
 *	Contains three routines that drive the body of the work, one for
 *	each level of readmes -- the top level, the category level, and
 *	the individual port level -- plus a few small auxiliary functions
 *
 *	main() will call only one of the make_readme_* routines, based on whether
 *	the pathname given as a command line argument (or the top-level ports
 *	directory, if none is provided) is determined to be at the top, category,
 *	or individual port level
 *
 *	$Id: mkreadmes.c,v 1.90 2012/03/05 03:59:59 conrads Exp $
 *
 *****************************************************************************/

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <unistd.h>

#include "mkreadmes.h"

/* placeholder names used in templates */

static char *placeholder[] = {
	"%%PKG%%",
	"%%PORT%%",
	"%%PREFIX%%",
	"%%COMMENT%%",
	"%%DESCR%%",
	"%%EMAIL%%",
	"%%CATEGORY%%",
	"%%BUILD_DEPENDS%%",
	"%%RUN_DEPENDS%%",
	"%%WEBSITE%%",
	"%%TOP%%",
	"%%SUBDIR%%",
	NULL			/* end of list marker */
};

static	char	ports[MAXPORTSLIST];	/* list of ports found in a category's
											Makefile
										 */
static	char	category_comment[LINE_MAX];	/* comment from a category's Makefile
											 */
/* pointers to dynamically allocated buffers for the template files */
char	*template_buf_category = NULL;
char	*template_buf_port = NULL;
char	*template_buf_top = NULL;

extern	char	ports_dir[PATH_MAX];		/* top-level ports directory */

extern	char	categories[];		/* list of categories found in the top-level
										Makefile
									 */
extern	int		num_categories;		/* number of categories found */
extern	char	*category;			/* pointer to current category */

extern	char	**index_categories;	/* pointer to an array of pointers to the first
										line of each category's section within the
										index buffer
									 */
extern	char	*index_field[];	/* pointers to the current port's index fields */

extern	char	*next_index;	/* pointer to the next index line to search */
extern	char	*next_cat;	/* pointer to the first index line of next category
							 */

extern	int		quiet;			/* silence normal output */
extern	int		verbose;		/* increase the amount of output */

/* prototypes for routines in other source files */

extern	int		search_index(const char *path);					/* index.c */
extern	char	*skip_leading(const char *str, const char *pattern);/* init.c */

/*****************************************************************************/

static char *HTMLify(char *ptr, char *str )
/*
	Copy 'str' to 'ptr', replacing any occurences of '&', '<' or '>'
	with their HTML	representations

	Returns a pointer to the location in 'ptr' one past the last
	character copied

	Uses the addstr() macro defined in mkreadmes.h
 */
{
	while (*str)
	{
		switch(*str)
		{
			case '&':
			{
				addstr(ptr, "&amp;");
				++str;
				continue;
			}
			case '<':
			{
				addstr(ptr, "&lt;");
				++str;
				continue;
			}
			case '>':
			{
				addstr(ptr, "&gt;");
				++str;
				continue;
			}
			default:
				*ptr++ = *str++;
		}
	}

	return ptr;
}

/*****************************************************************************/

static int match_placeholder(const char *ptr)
/*
	Try to find a matching "%%VARIABLE%%" within the placeholder array
	for the string pointed to by 'ptr'

	On entry to this routine, 'ptr' will be pointing to the first "%" in the
	possible symbol name we're looking for, so all we have to do is see if the
	next character is also a "%" and then scan the placeholder array for a match

	Returns a positive integer representing the variable's index in the
	placeholder array, or -1 if no match is found
 */
{
	if (ptr[1] == '%')
	{
		int i = 0;

		/* We've hit a double "%%", now find its match in the placeholder
			array, and return its index within the placeholder array
		 */

		while (placeholder[i])
		{
			if (strncmp(ptr, placeholder[i], strlen(placeholder[i])) == 0)
				return i;
			++i;
		}
	}

	/* No match found */
	return -1;
}

/*****************************************************************************/

static int category_index(char *category)
/*
	scan the list of category names to get an index to use with the
	index_categories[] array to get the pointer to the first index line of
	category in the index buffer

	returns the index at which the category was found, or num_categories
	if not found
*/
{
	char   *categories_ptr = categories;
	int		i = 0;

	while (*categories_ptr)
	{
		if (strcmp(category, categories_ptr) == 0)
			return i;									/* found a match */
		/* else */
		categories_ptr += strlen(categories_ptr) + 1;	/* bump pointer and */
		++i;											/* index */
	}

	return i;	/* (i == num_categories) = not found */
}

/*****************************************************************************/

static int make_readme_port(const char *path, int make_level)
/*
	Make the README.html file for a single port

	On entry, category will already be pointing to the name of the port's
	category

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_PORT];	/* output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	int		readme;	/* file descriptor for the README.html we're creating */

	char   *template_ptr = template_buf_port;	/* pointer into template
													buffer
												 */
	char	port_category[MAXCATNAME];	/* category for this port */
	int		catlen;						/* length of category name */

	int		i;				/* index into the placeholder array where a
								template match was found
							 */

	if (make_level == PORT)	/* need to set category, next_index and next_cat */
	{
		/* on entry, category points to the first character after "ports_dir/"
			in path, so copy everything up to the next slash to port_category
		 */

		catlen = strcspn(category, "/");
		strncpy(port_category, category, catlen);
		port_category[catlen] = '\0';
		category = port_category;	/* reset category to point to the string we
										just copied
									 */

		/* set next_index to the start of this port's category's section
			in the index buffer
		 */
		next_index = index_categories[category_index(category)];

		if (next_index == NULL)	/* category not found :-( */
		{
			fprintf(stderr, "No such category: %s\n", category);
			return 1;
		}

		/* get a pointer to the first index line of the next category
			from the index_categories[] array */
		next_cat = index_categories[category_index(category) + 1];
	}

	/* Find the port's entry in the index file, saving its
		info in the index_field[] array
	 */
	if (!search_index(path))
	{
		fprintf(stderr, "Port %s not found in INDEX\n", path);
		return 1;
	}

	/* For efficiency's sake, don't output the following message
		unless this one port is all we're building, or the verbose
		flag is on
	 */
	if (verbose || ((make_level == PORT) && !quiet))
		printf("Making README.html for port %s\n",
			skip_leading(path, ports_dir));

	/* Open the port's README.html file for writing */

	strcpy(readme_path, path);
	strcat(readme_path, "/");
	strcat(readme_path, READMENAME);

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* a few fields require special handling:

				for BUILD_DEPENDS and RUN_DEPENDS, don't output anything
				if these fields are empty in the INDEX file, otherwise,
				precede them with the introductory text #defined in
				BUILD_DEPS or RUN_DEPS

				for WEBSITE, output the lead-in text #defined in WWW_BEFORE,
				then the field contents, and follow it up with the text
				#defined in WWW_AFTER

				for the PORT field, chop off the leading ports_dir/ from the
				pathname and just use category/port in the output

				for the COMMENT field, convert any '&', '<' or '>' characters
				into their HTML representations
			 */
			switch (i)
			{
				case PKG:
					addstr(readme_ptr, index_field[i]);
					break;
				case DESCR:
					addstr(readme_ptr, ports_dir);
					addstr(readme_ptr, skip_leading(index_field[i], PORTSTOP));
					break;
				case EMAIL:
					addstr(readme_ptr, index_field[i]);
					break;
				case COMMENT:
					/* the port's comment may have some bad HTML chars */
					readme_ptr = HTMLify(readme_ptr, index_field[COMMENT]);
					break;
				case WEBSITE:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, WWW_BEFORE);
						addstr(readme_ptr, index_field[i]);
						addstr(readme_ptr, WWW_AFTER);
					}
					break;
				case PORT:
					addstr(readme_ptr, skip_leading(index_field[i], PORTSTOP));
					break;
				case BUILD_DEPENDS:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, BUILD_DEPS);
						addstr(readme_ptr, index_field[i]);
					}
					break;
				case RUN_DEPENDS:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, RUN_DEPS);
						addstr(readme_ptr, index_field[i]);
					}
					break;
				case TOP:
					addstr(readme_ptr, ports_dir);
					break;
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

static int make_readme_category(const char *path, int make_level)
/*
	Make the README.html file for a single category

	Calls make_readme_port() to build the README.html file for each
	port within the category

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_CAT];	/* output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	int		readme;	/* file descriptor for the README.html we're creating */

	char   *template_ptr = template_buf_category;	/* pointer into template
														buffer
													 */

	char	Makefile_buf[LINE_MAX];		/* input buffer */
	char	Makefile_path[PATH_MAX];	/* input pathname */
	FILE   *Makefile;					/* input stream for category's
											Makefile
										 */

	char   *ports_ptr = ports;	/* pointer into ports array */
	char   *port;				/* pointer to port in current input buffer */
	char	port_path[PATH_MAX];

	char   *comment_ptr = Makefile_buf;

	int		i;					/* index into the placeholder array where a
									template match was found
								 */
	int		num_ports = 0;

	/* set next_index to the start of this category's section
		in the index buffer
	 */
	next_index = index_categories[category_index(category)];

	if (next_index == NULL)	/* category not found :-( */
	{
		fprintf(stderr, "No such category: %s\n", category);
		return 1;
	}

	/* get a pointer to the first index line of the next category from the
		index_categories[] array */
	next_cat = index_categories[category_index(category) + 1];

	/* generate the pathname to the category's Makefile */

	strcpy(Makefile_path, path);
	strcat(Makefile_path, "/");
	strcat(Makefile_path, MAKEFILENAME);

	/* Open the Makefile for reading */

	if ((Makefile = fopen(Makefile_path, "r")) == NULL)
	{
		perror(Makefile_path);
		return 1;
	}

	/* First, get the category COMMENT, saving it in the external buffer
		category_comment[], in case make_readme_top() needs it
	 */

	while ((fgets(Makefile_buf, LINE_MAX, Makefile) != NULL) &&
			(comment_ptr == Makefile_buf))
	{
		if ((comment_ptr = skip_leading(Makefile_buf, "    COMMENT = "))
			!= Makefile_buf)
		{
			/* don't want the newline */
			comment_ptr[strlen(comment_ptr) - 1] = '\0';
			strcpy(category_comment, comment_ptr);
			break;
		}
	}

	/* Got the comment, continue reading to get all the SUBDIRS */

	while ((fgets(Makefile_buf, LINE_MAX, Makefile) != NULL))
	{
		if ((port = skip_leading(Makefile_buf, "    SUBDIR += ")) !=
				Makefile_buf)	/* found a SUBDIR line */
		{
			/* don't want the newline */
			port[strlen(port) - 1] = '\0';
			ports_ptr = stpcpy(ports_ptr, port) + 1;
			++num_ports;
		}
	}

	/* add an extra nul character to mark the end of the ports array */
	*ports_ptr = '\0';

	fclose(Makefile);

	/* Done collecting the basic data, now we can start generating
		the README.html file
	 */
	if (!quiet)
		printf("Making README.html for category %s (%d ports)\n",
			skip_leading(path, ports_dir), num_ports);

	/* Open the category's README.html file for writing */

	strcpy(readme_path, path);
	strcat(readme_path, "/");
	strcat(readme_path, READMENAME);

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* SUBDIR requires special handling, calling make_readme_port()
				for each port in the category to build the port's README.html
				file and to obtain its package name and comment for the
				category README.html file

				same as in make_readme_port(), for the COMMENT field, convert
				any '&', '<' or '>' characters into their HTML representations
			 */
			switch (i)
			{
				case COMMENT:
					addstr(readme_ptr, category_comment);
					/* add the number of ports in category to end of comment */
					readme_ptr += sprintf(readme_ptr, " (%d ports)", num_ports);
					break;
				case CATEGORY:
					addstr(readme_ptr, category);
					break;
				case SUBDIR:
					ports_ptr = ports;

					while (*ports_ptr)
					{
						port = ports_ptr;	/* port's base name */

						/* generate the full path to the port */
						strcpy(port_path, path);
						strcat(port_path, "/");
						strcat(port_path, port);

						/* make the port's README.html */
						make_readme_port(port_path, make_level);

						/* now we can use the information left over in the
							index_field array */
						addstr(readme_ptr, "<a href=\"");
						addstr(readme_ptr, port);
						addstr(readme_ptr, "/README.html\">");
						addstr(readme_ptr, index_field[PKG]);
						addstr(readme_ptr, "</a>: ");

						/* the port's comment may have some bad HTML chars */
						readme_ptr = HTMLify(readme_ptr, index_field[COMMENT]);
						strcpy(readme_ptr++, "\n");

						/* bump ports_ptr to next item in ports array */
						ports_ptr += strlen(port) + 1;
					}
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

static int make_readme_top(const char *path, int make_level)
/*
	Make the top-level README.html file

	Calls make_readme_category() to build the README.html file for each
	category in the categories list created earlier from the top-level Makefile

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_TOP];	/* output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	int		readme;	/* file descriptor for the README.html we're creating */

	char   *template_ptr = template_buf_top;	/* pointer into template buffer
												 */
	char   *categories_ptr = categories;	/* pointer into categories list */

	char	category_path[PATH_MAX];	/* fully qualified path to current
											category
										 */
	int		i;						/* index into the placeholder array where
										a template match was found
									 */

	if (!quiet)
		printf("Making README.html for %s (%d categories)\n", ports_dir,
			num_categories);

	/* Open the top-level README.html file for writing */

	strcpy(readme_path, path);
	strcat(readme_path, READMENAME);

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* %%SUBDIR%% is the only placeholder used in the
				top-level template

				For each category found in the top-level Makefile, do a
				make_readme_category() to obtain its comment
			 */
			if (i == SUBDIR)
			{
				categories_ptr = categories;

				while (*categories_ptr)
				{
					category = categories_ptr;	/* category's base name */

					/* generate the full path to the category */
					strcpy(category_path, ports_dir);
					strcat(category_path, category);

					/* make the category's README.html, saving its
						COMMENT in category_comment[] */
					make_readme_category(category_path, make_level);

					/* now we have all the information we need
						for this category
					 */
					addstr(readme_ptr, "<a href=\"");
					addstr(readme_ptr, category);
					addstr(readme_ptr, "/README.html\">");
					addstr(readme_ptr, category);
					addstr(readme_ptr, "</a>: ");
					addstr(readme_ptr, category_comment);
					strcpy(readme_ptr++, "\n");

					/* bump pointer to next item in categories array */
					categories_ptr += strlen(category) + 1;
				}
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

int make_readme(const char *path, int make_level)
/*
	wrapper for the actual make_readme_* routines, called from main()

	returns the return code of whichever routine it calls

	(the default case should never happen, and is provided merely to avoid
	having the compiler generate a warning at compile time)
 */
{
	switch (make_level)
	{
		case CATEGORY:	return make_readme_category(path, make_level);
		case PORT:		return make_readme_port(path, make_level);
		case TOP:		return make_readme_top(path, make_level);
		default:		return -1;
	}
}

